""" Tests the dictionary based FC evaluator"""
from typing import Optional
from unittest import mock

import pytest  # type:ignore[import]

from ahbicht.content_evaluation import fc_evaluators
from ahbicht.content_evaluation.fc_evaluators import DictBasedFcEvaluator, FcEvaluator
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint


class TestDictBasedFcEvaluator:
    """Test for the evaluation using the Dict Based FC Evaluator"""

    hardcoded_results = {
        "1": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
        "2": EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something wrong"),
    }

    @pytest.fixture()
    def dict_fc_evaluator(self) -> FcEvaluator:
        evaluator = DictBasedFcEvaluator(self.hardcoded_results)
        return evaluator

    @pytest.mark.parametrize(
        "condition_key,text_input, expected_result",
        [
            pytest.param("1", "asd", EvaluatedFormatConstraint(format_constraint_fulfilled=True)),
            pytest.param(
                "2",
                "yxc",
                EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something wrong"),
            ),
        ],
    )
    async def test_evaluation(
        self,
        condition_key: str,
        text_input: Optional[str],
        expected_result: EvaluatedFormatConstraint,
        dict_fc_evaluator,
    ):
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set("asd")
        assert await dict_fc_evaluator.evaluate_single_format_constraint("1") == EvaluatedFormatConstraint(
            format_constraint_fulfilled=True
        )
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set("yxc")
        assert await dict_fc_evaluator.evaluate_single_format_constraint("2") == EvaluatedFormatConstraint(
            format_constraint_fulfilled=False, error_message="something wrong"
        )

    async def test_not_implemented(self, dict_fc_evaluator):
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set("qwe")
        with pytest.raises(NotImplementedError):
            await dict_fc_evaluator.evaluate_single_format_constraint("3")

    async def test_multithreading_with_same_value(self, dict_fc_evaluator, mocker):
        dict_evaluation_spy = mocker.spy(dict_fc_evaluator, "evaluate_single_format_constraint")
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set("asd")
        assert await dict_fc_evaluator.evaluate_format_constraints(["1", "2"]) == self.hardcoded_results
        dict_evaluation_spy.assert_has_awaits([mock.call("1"), mock.call("2")])



import asyncio
from itertools import product
from typing import List

import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]
from lark import Tree

from ahbicht.expressions.ahb_expression_parser import _parser as ahb_expr_parser
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import _parser as cond_expr_parser
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree


class TestCaching:
    """
    Tests the caching capabilities of both ahb and condition expression parsers
    """

    def test_ahb_expression_cache_sync(self, mocker):
        ahb_expression = "Muss [3] U [4]"
        parse_spy = mocker.spy(ahb_expr_parser, "parse")
        tree_instances: List[Tree] = []
        number_of_calls: int = 100
        for _ in range(number_of_calls):
            tree_instance = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
            tree_instances.append(tree_instance)
        parse_spy.assert_called_once_with(ahb_expression)
        # The following assertion is to make sure that each cached call actually returns a new instance of a lark tree.
        # We do not want the same instance to be returned and eventually be modified over and over again.
        number_of_distinct_instances: int = (
            len(tree_instances)
            * len(tree_instances)
            / len([1 for x, y in product(tree_instances, tree_instances) if x is y])
        )
        assert number_of_distinct_instances == number_of_calls

    async def test_ahb_expression_cache_async(self, mocker):

        ahb_expression = "Muss [5] U [6]"
        parse_spy = mocker.spy(ahb_expr_parser, "parse")

        async def parsing_task():
            parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)

        tasks = [parsing_task() for _ in range(100)]
        await asyncio.gather(*tasks)
        parse_spy.assert_called_once_with(ahb_expression)

    def test_condition_expression_cache_sync(self, mocker):
        cond_expression = "[1] U [2]"
        parse_spy = mocker.spy(cond_expr_parser, "parse")
        tree_instances: List[Tree] = []
        number_of_calls: int = 100
        for _ in range(number_of_calls):
            tree_instance = parse_condition_expression_to_tree(cond_expression)
            tree_instances.append(tree_instance)
        parse_spy.assert_called_once_with(cond_expression)
        number_of_distinct_instances: int = (
            len(tree_instances)
            * len(tree_instances)
            / len([1 for x, y in product(tree_instances, tree_instances) if x is y])
        )
        assert number_of_distinct_instances == number_of_calls

    async def test_condition_expression_cache_async(self, mocker):
        # test expression has to be different from the cond_expression in the sync test case because the tests interfere
        cond_expression = "[3] U [4]"
        parse_spy = mocker.spy(cond_expr_parser, "parse")

        async def parsing_task():
            parse_condition_expression_to_tree(cond_expression)

        tasks = [parsing_task() for _ in range(100)]
        await asyncio.gather(*tasks)
        parse_spy.assert_called_once_with(cond_expression)



"""This module contains the enums for the possible validation values."""

from enum import Enum


class RequirementValidationValue(str, Enum):
    """
    Possible values to describe the state of the validation
    in the requirement_validation attribute of the ValidationResult.
    The values without "AND_EMPTY" or "AND_FILLED" is for segment_level
    due to no value that could be filled or empty.
    It's easier to put all of them in one Enum for the ValidationResult class.
    """

    IS_REQUIRED = "IS_REQUIRED"  #: element is required
    IS_FORBIDDEN = "IS_FORBIDDEN"  #: element is forbidden
    IS_OPTIONAL = "IS_OPTIONAL"  #: element is optional
    IS_REQUIRED_AND_EMPTY = "IS_REQUIRED_AND_EMPTY"  #: field is required, but empty
    IS_REQUIRED_AND_FILLED = "IS_REQUIRED_AND_FILLED"  #: field is required and filled
    IS_FORBIDDEN_AND_EMPTY = "IS_FORBIDDEN_AND_EMPTY"  #: field is forbidden and empty
    IS_FORBIDDEN_AND_FILLED = "IS_FORBIDDEN_AND_FILLED"  #: field is forbidden, but filled
    IS_OPTIONAL_AND_EMPTY = "IS_OPTIONAL_AND_EMPTY"  #: field is optional and empty
    IS_OPTIONAL_AND_FILLED = "IS_OPTIONAL_AND_FILLED"  #: field is optional and filled

    def __str__(self):
        return self.value



"""
This module parses a condition expression like "[59] U ([123] O [456])" into a tree structur using
the parsing library lark: https://lark-parser.readthedocs.io/en/latest/

The used terms are defined in the README_conditions.md.
"""
# pylint:disable=cyclic-import
from functools import lru_cache
from typing import List, Union

from lark import Lark, Token, Tree
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF

from ahbicht.condition_node_distinction import ConditionNodeType, derive_condition_node_type
from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtract
from ahbicht.expressions import parsing_logger
from ahbicht.utility_functions import tree_copy

GRAMMAR = r"""
?expression: expression "O"i expression -> or_composition
            | expression "∨" expression -> or_composition
            | expression "X"i expression -> xor_composition
            | expression "⊻" expression -> xor_composition
            | expression "U"i expression -> and_composition
            | expression "∧" expression -> and_composition
            | expression expression -> then_also_composition
            | brackets
            | package
            | condition
            | time_condition
?brackets: "(" expression ")"
time_condition: "[" TIME_CONDITION_KEY "]" // a rule for point in time-conditions
package: "[" PACKAGE_KEY REPEATABILITY? "]" // a rule for packages
condition: "[" CONDITION_KEY "]" // a rule for condition keys
TIME_CONDITION_KEY: /UB(1|2|3)/ // a terminal for "übergreifende Bedingungen für Zeitpunktangaben"
CONDITION_KEY: INT // a TERMINAL for all the remaining ints (lower priority)
REPEATABILITY: /\d+\.{2}[1-9]\d*/ // a terminal for repetitions n..m with n>=0 and m>n
PACKAGE_KEY: INT "P" // a TERMINAL for all INTs followed by "P" (high priority)
%import common.INT
%import common.WS
%ignore WS  // WS = whitespace
"""
_parser = Lark(GRAMMAR, start="expression")


@tree_copy
@lru_cache(maxsize=65536)
def parse_condition_expression_to_tree(condition_expression: str) -> Tree[Token]:
    """
    Parse a given condition expression with the help of the here defined grammar to a lark tree.
    The grammar starts with condition keys, e.g. [45] and combines them with
    and _/or_compositions corresponding to U/O operators or without an operator (then_also_composition).
    It follows the boolean logic 'brackets before `then_also` before `and` before `xor` before `or`'.
    Whitespaces are ignored.

    :param condition_expression: str, e.g. '[45]U[502]O[1][906]'
    :return parsed_tree: Tree
    """
    try:
        parsed_tree = _parser.parse(condition_expression)
        parsing_logger.debug("Successfully parsed '%s' as condition expression", condition_expression)
    except (UnexpectedEOF, UnexpectedCharacters, TypeError) as eof:
        raise SyntaxError(
            """Please make sure that:
             * all conditions have the form [INT]
             * all packages have the form [INTPn..m]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """
        ) from eof
    return parsed_tree


def extract_categorized_keys_from_tree(
    tree_or_list: Union[Tree, List[str]], sanitize: bool = False
) -> CategorizedKeyExtract:
    """
    find different types of condition nodes inside the given tree or list of keys.
    The types are differentiated by their number range.
    See 'Allgemeine Festlegungen' from EDI@Energy.
    """
    result = CategorizedKeyExtract(
        format_constraint_keys=[], requirement_constraint_keys=[], hint_keys=[], package_keys=[], time_condition_keys=[]
    )
    condition_keys: List[str]
    if isinstance(tree_or_list, list):
        condition_keys = tree_or_list
    elif isinstance(tree_or_list, Tree):
        condition_keys = [
            x.value  # type:ignore[attr-defined]
            for x in tree_or_list.scan_values(
                lambda token: token.type == "CONDITION_KEY"  # type:ignore[union-attr]
            )
        ]
        result.package_keys = [
            x.value  # type:ignore[attr-defined]
            for x in tree_or_list.scan_values(
                lambda token: token.type == "PACKAGE_KEY"  # type:ignore[union-attr]
            )
        ]
        result.time_condition_keys = [
            x.value  # type:ignore[attr-defined]
            for x in tree_or_list.scan_values(
                lambda token: token.type == "TIME_CONDITION_KEY"  # type:ignore[union-attr]
            )
        ]
    else:
        raise ValueError(f"{tree_or_list} is neither a list nor a {Tree.__name__}")
    for condition_key in condition_keys:
        condition_node_type = derive_condition_node_type(condition_key)
        if (
            condition_node_type is ConditionNodeType.REQUIREMENT_CONSTRAINT
            or condition_node_type is ConditionNodeType.REPEATABILITY_CONSTRAINT
        ):
            result.requirement_constraint_keys.append(condition_key)
        elif condition_node_type is ConditionNodeType.HINT:
            result.hint_keys.append(condition_key)
        elif condition_node_type is ConditionNodeType.FORMAT_CONSTRAINT:
            result.format_constraint_keys.append(condition_key)
        else:
            # if you run into the ConditionNodeType.PACKAGE case, you probably forgot to enable the package resolving
            # in the parsing step (although the error occurs during evaluation)
            raise NotImplementedError(f"The type '{condition_node_type}' is not implemented yet.")
    if sanitize:
        result.sanitize()
    return result


async def extract_categorized_keys(
    condition_expression: str, resolve_packages: bool = False, replace_time_conditions: bool = False
) -> CategorizedKeyExtract:
    """
    Parses the given condition expression and returns CategorizedKeyExtract as a template for content
    evaluation.
    """
    # because of
    # ImportError: cannot import name 'parse_condition_expression_to_tree' from partially initialized module
    # 'ahbicht.expressions.condition_expression_parser' (most likely due to a circular import)
    # pylint: disable=import-outside-toplevel
    from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions

    tree = await parse_expression_including_unresolved_subexpressions(
        condition_expression, resolve_packages=resolve_packages, replace_time_conditions=replace_time_conditions
    )
    return extract_categorized_keys_from_tree(tree, sanitize=True)



""" Tests the dictionary based RC evaluator"""
from unittest import mock

import pytest  # type:ignore[import]

from ahbicht.content_evaluation.rc_evaluators import DictBasedRcEvaluator
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue
from unittests.defaults import empty_default_test_data


class TestDictBasedRcEvaluator:
    """Test for the evaluation using the Dict Based RC Evaluator"""

    async def test_evaluation(self, mocker):
        hardcoded_results = {
            "1": ConditionFulfilledValue.NEUTRAL,
            "2": ConditionFulfilledValue.UNFULFILLED,
            "3": ConditionFulfilledValue.FULFILLED,
            "4": ConditionFulfilledValue.UNKNOWN,
        }
        evaluator = DictBasedRcEvaluator(hardcoded_results)
        dummy_eval_data = empty_default_test_data
        assert (
            await evaluator.evaluate_single_condition("1", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.NEUTRAL
        )
        assert (
            await evaluator.evaluate_single_condition("2", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.UNFULFILLED
        )
        assert (
            await evaluator.evaluate_single_condition("3", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.FULFILLED
        )
        assert (
            await evaluator.evaluate_single_condition("4", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.UNKNOWN
        )
        with pytest.raises(NotImplementedError):
            await evaluator.evaluate_single_condition("5", evaluatable_data=dummy_eval_data)

        single_condition_spy = mocker.spy(evaluator, "evaluate_single_condition")

        assert (
            await evaluator.evaluate_conditions(["1", "2", "3", "4"], evaluatable_data=dummy_eval_data)
            == hardcoded_results
        )
        single_condition_spy.assert_has_awaits(
            [
                mock.call("1", dummy_eval_data, None),
                mock.call("2", dummy_eval_data, None),
                mock.call("3", dummy_eval_data, None),
                mock.call("4", dummy_eval_data, None),
            ]
        )



import asyncio
from contextvars import ContextVar
from logging import LogRecord
from typing import List

import inject
import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]
from maus import (
    DataElementFreeText,
    DataElementValuePool,
    DeepAnwendungshandbuch,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)
from maus.models.anwendungshandbuch import AhbMetaInformation

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.validation.validation import validate_deep_anwendungshandbuch
from unittests.defaults import (
    DefaultHintsProvider,
    DefaultPackageResolver,
    EmptyDefaultFcEvaluator,
    EmptyDefaultRcEvaluator,
    default_test_format,
    default_test_version,
)

pytestmark = pytest.mark.asyncio


class SomethingInjectable:
    """
    This class is used by one of our custom Evaluators.
    The evaluators a bit fragile because of https://github.com/ivankorobkov/python-inject/issues/77
    This means you _cannot_ use inject.attr(...) in your custom evaluators
    """

    pass


class MweRcEvaluator(EmptyDefaultRcEvaluator):
    # do _not_use inject.attr! It breaks in the inspection inside the Evaluator __init__ method

    def evaluate_1(self, evaluatable_data, context):
        seed = evaluatable_data.edifact_seed
        if "foo" in seed and seed["foo"] == "bar":
            return ConditionFulfilledValue.FULFILLED
        return ConditionFulfilledValue.UNFULFILLED

    async def evaluate_2(self, evaluatable_data, context):
        # an async method, just for the sake of it
        seed = evaluatable_data.edifact_seed
        if "asd" in seed and seed["asd"] == "yxc":
            return ConditionFulfilledValue.FULFILLED
        return ConditionFulfilledValue.UNFULFILLED

    def evaluate_3(self, _, __):
        something_injected = inject.instance(SomethingInjectable)  # this is fine (other than inject.attr, see above)
        assert isinstance(something_injected, SomethingInjectable)
        return ConditionFulfilledValue.UNFULFILLED


class MweFcEvaluator(EmptyDefaultFcEvaluator):
    def evaluate_998(self, entered_input):
        """fantasy FC: input must be all upper case"""
        if not entered_input:
            return EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="Input must not be empty")
        if entered_input.to_upper() == entered_input:
            return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)
        return EvaluatedFormatConstraint(
            format_constraint_fulfilled=False, error_message=f"Input '{entered_input}' is not all upper case"
        )


current_evaluatable_data: ContextVar[EvaluatableData] = ContextVar("current_evaluatable_data")


def get_current_evaluatable_data() -> EvaluatableData:
    return current_evaluatable_data.get()


class TestIntegrationMwe:
    """
    Contains an integration tests that show a full minimal working example (meaning: no mocks at all).
    If any tests break, then first fix all other tests and run these tests last.
    """

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector(self):
        fc_evaluator = MweFcEvaluator()
        rc_evaluator = MweRcEvaluator()
        hints_provider = DefaultHintsProvider({"567": "Hallo Welt"})
        package_resolver = DefaultPackageResolver({"4P": "[1] U [2]"})
        inject.clear_and_configure(
            lambda binder: binder.bind(
                TokenLogicProvider,
                SingletonTokenLogicProvider([fc_evaluator, rc_evaluator, hints_provider, package_resolver]),
            ).bind_to_provider(EvaluatableDataProvider, get_current_evaluatable_data)
        )
        yield
        inject.clear()

    async def test_integration(self, setup_and_teardown_injector, caplog):
        maus = DeepAnwendungshandbuch(
            meta=AhbMetaInformation(pruefidentifikator="12345"),
            lines=[
                SegmentGroup(
                    discriminator="root",
                    ahb_expression="X[4P]",
                    segments=[
                        Segment(
                            discriminator="UNH",
                            ahb_expression="X",
                            data_elements=[
                                DataElementFreeText(
                                    discriminator="Nachrichten-Startsegment",
                                    ahb_expression="X[998][567]",
                                    entered_input=None,
                                    data_element_id="1234",
                                )
                            ],
                        )
                    ],
                ),
                SegmentGroup(
                    discriminator="SG4",
                    ahb_expression="X",
                    segments=[
                        Segment(
                            discriminator="FOO",
                            ahb_expression="X",
                            data_elements=[
                                DataElementValuePool(
                                    discriminator="SG4->FOO->0333",
                                    value_pool=[
                                        ValuePoolEntry(
                                            qualifier="E01",
                                            meaning="Das andere",
                                            ahb_expression="X[4P]",
                                        ),
                                        ValuePoolEntry(
                                            qualifier="E02",
                                            meaning="Das Eine",
                                            ahb_expression="X[3]",
                                        ),
                                    ],
                                    data_element_id="0333",
                                    entered_input=None,
                                )
                            ],
                        )
                    ],
                    segment_groups=[
                        SegmentGroup(
                            discriminator="SG5",
                            ahb_expression="X[2]",
                            segments=[
                                Segment(
                                    discriminator="BAR",
                                    ahb_expression="X",
                                    data_elements=[
                                        DataElementFreeText(
                                            discriminator="Die fünfte Gruppe",
                                            ahb_expression="X",
                                            entered_input=None,
                                            data_element_id="1234",
                                        )
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )

        async def first_evaluation():
            current_evaluatable_data.set(
                EvaluatableData(
                    edifact_seed={"foo": "bar", "asd": "yxc"},
                    edifact_format=default_test_format,
                    edifact_format_version=default_test_version,
                )
            )
            return await validate_deep_anwendungshandbuch(maus)

        async def second_evaluation():
            current_evaluatable_data.set(
                EvaluatableData(
                    edifact_seed={"foo": "baz", "asd": "qwe"},
                    edifact_format=default_test_format,
                    edifact_format_version=default_test_version,
                )
            )  # use a different message under test to trigger different outcomes
            # you can set a breakpoint in evaluate_1 and evaluate_2 to manually verify the different data they access
            return await validate_deep_anwendungshandbuch(maus)

        results1, results2 = await asyncio.gather(*[first_evaluation(), second_evaluation()])
        assert results2 is not None
        assert results1 != results2  # this shows, that the evaluatable data used are indeed different in each call
        log_entries: List[LogRecord] = caplog.records
        assert len([le for le in log_entries if le.message.startswith("The requirement constraint")]) == 12
        assert len([le for le in log_entries if le.message.startswith("The format constraint")]) == 1
